open Prog
open Elang
open Elang_run
open Batteries
open BatList
open Cfg
open Utils
open Builtins

let rec eval_cfgexpr oc (cp : cprog) (st: int state) (e: expr) : int res =
  match e with
  | Ebinop(b, e1, e2) ->
    eval_cfgexpr oc cp st e1 >>= fun v1 ->
    eval_cfgexpr oc cp st e2 >>= fun v2 ->
    let v = eval_binop b v1 v2 in
    OK v
  | Eunop(u, e) ->
    eval_cfgexpr oc cp st e >>= fun v1 ->
    let v = (eval_unop u v1) in
    OK v
  | Eint i -> OK i
  | Evar s ->
    begin match Hashtbl.find_option st.env s with
      | Some v -> OK v
      | None -> Error (Printf.sprintf "Unknown variable %s\n" s)
    end

  | Ecall(fname,expr_args) ->  list_map_res (fun eXpr -> eval_cfgexpr oc cp st eXpr) expr_args >>= 
    fun vargs ->  match find_function cp fname with
    |OK(f) -> eval_cfgfun oc cp st fname f vargs >>= 
          fun (ret,st') ->  option_to_res_bind ret "no return" (fun ret -> OK(ret))
    |Error(s) ->  do_builtin oc st.mem fname vargs >>= 
          fun r -> match r with
          |None -> Error (Printf.sprintf "%s ne retourne rien " fname)
          |Some(i) -> OK(i)

and eval_cfginstr oc (cp : cprog) st ht (n: int): (int * int state) res =
  match Hashtbl.find_option ht n with
  | None -> Error (Printf.sprintf "Invalid node identifier\n")
  | Some node ->
    match node with
    | Cnop succ ->
      eval_cfginstr oc cp st ht succ
    | Cassign(v, e, succ) ->
      eval_cfgexpr oc cp st e >>= fun i ->
      Hashtbl.replace st.env v i;
      eval_cfginstr oc cp st ht succ
    | Ccmp(cond, i1, i2) ->
      eval_cfgexpr oc cp st cond >>= fun i ->
      if i = 0 then eval_cfginstr oc cp st ht i2 else eval_cfginstr oc cp st ht i1
    | Creturn(e) ->
      eval_cfgexpr oc cp st e >>= fun e ->
      OK (e, st)
    | Cprint(e, succ) ->
      eval_cfgexpr oc cp st e >>= fun e ->
      Format.fprintf oc "%d\n" e;
      eval_cfginstr oc cp st ht succ
    | Ccall(fname,expr_args,succ) ->  list_map_res (fun eXpr -> eval_cfgexpr oc cp st eXpr) expr_args >>= 
        fun vargs ->  match find_function cp fname with
        |OK(f) -> eval_cfgfun oc cp st fname f vargs >>= 
              fun (ret,st') ->  eval_cfginstr oc cp {st with mem = st'.mem} ht succ
        |Error(s) ->  do_builtin oc st.mem fname vargs >>= 
              fun r ->  eval_cfginstr oc cp (st) ht succ
     

and eval_cfgfun oc ep st cfgfunname { cfgfunargs;
                                      cfgfunbody;
                                      cfgentry} vargs =
  let st' = { st with env = Hashtbl.create 17 } in
  match List.iter2 (fun a v -> Hashtbl.replace st'.env a v) cfgfunargs vargs with
  | () -> eval_cfginstr oc ep st' cfgfunbody cfgentry >>= fun (v, st') ->
    OK (Some v, {st' with env = st.env})
  | exception Invalid_argument _ ->
    Error (Format.sprintf "CFG: Called function %s with %d arguments, expected %d.\n"
             cfgfunname (List.length vargs) (List.length cfgfunargs)
          )

let eval_cfgprog oc cp memsize params =
  let st = init_state memsize in
  find_function cp "main" >>= fun f ->
  let n = List.length f.cfgfunargs in
  let params = take n params in
  eval_cfgfun oc cp st "main" f params >>= fun (v, st) ->
  OK v


